"""Test .tab format functionalities"""
import pathlib
import tempfile

import numpy as np

import afmformats


datadir = pathlib.Path(__file__).resolve().parent / "data"


def generate_tab_file():
    tf = tempfile.mktemp(suffix=".tab", prefix="afmformats_test_tab_")
    with open(tf, "w") as fd:
        fd.write("# force\ttip position\tsegment\r\n")
        for ii in range(100):
            fd.write("{}\t{}\t{}\r\n".format(ii*1e-9,
                                             np.pi*ii*1e-6,
                                             ii > 50))
    return tf


def test_open_simple():
    tf = generate_tab_file()
    data = afmformats.load_data(tf)[0]
    assert np.sum(data["segment"]) == 49
    assert np.allclose(data["force"][0], 0)
    assert np.allclose(data["force"][1], 1e-9)
    assert np.allclose(data["tip position"][1], np.pi*1e-6)
    # cleanup
    try:
        pathlib.Path(tf).unlink()
    except OSError:
        pass


def test_save_open_with_metadata():
    jpkfile = datadir / "spot3-0192.jpk-force"
    fdat = afmformats.load_data(jpkfile, mode="force-distance")[0]
    _, path = tempfile.mkstemp(suffix=".tab", prefix="afmformats_test_")
    fdat.export(path, metadata=True, fmt="tab")
    fdat2 = afmformats.load_data(path, mode="force-distance")[0]
    # compare metadata
    for key in fdat.metadata:
        if key == "path":
            assert fdat.metadata["path"] != fdat2.metadata["path"]
        else:
            assert fdat.metadata[key] == fdat2.metadata[key]
    for col in fdat.columns:
        assert np.allclose(fdat[col], fdat2[col], atol=0)
    # cleanup
    try:
        pathlib.Path(path).unlink()
    except OSError:
        pass


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import json
import pathlib

import numpy as np

from .afm_data import column_dtypes, known_columns


__all__ = ["load_tab"]


def load_tab(path, callback=None):
    """Loads tab-separated-value files as exported by afmformats

    This is a simple tab-separated values files. The metadata
    may be present at the beginning of the file, commented out,
    as a json dump in a "BEGIN METADATA" - "END METADATA" block.
    The column data is listed below as a simple table.
    """
    path = pathlib.Path(path)
    with path.open() as fd:
        tsvdata = fd.readlines()

    # get the metadata
    dump = []
    injson = False
    for ii, line in enumerate(tsvdata):
        if line.startswith("# BEGIN METADATA"):
            injson = True
            continue
        elif line.startswith("# END METADATA"):
            break
        elif injson:
            dump.append(line.strip("#").strip())
    if dump:
        metadata = json.loads("\n".join(dump))
    else:
        metadata = {}
    metadata["path"] = path
    metadata["enum"] = 0

    # last line with a hash is the header
    for ii, line in enumerate(tsvdata):
        if not line.strip():
            # empty line
            pass
        elif line.startswith("#"):
            # header candidate
            header_line = line
        else:
            if ii == 0:
                raise ValueError("No header found in '{}'!".format(path))
            break
    else:
        raise ValueError("No data found in '{}'!".format(path))
    columns = header_line.strip("#").strip().split("\t")

    # load the data
    da = [f.strip() for f in tsvdata if f.strip() and not f.startswith("#")]
    # generate arrays
    data = {}
    for cc in columns:
        if cc in known_columns:
            data[cc] = np.zeros(len(da), dtype=column_dtypes[cc])
    for ii, line in enumerate(da):
        for jj, item in enumerate(line.strip().split("\t")):
            assert jj < len(columns)
            cc = columns[jj]
            if cc in known_columns:
                data[cc][ii] = string_to_dtype(item, column_dtypes[cc])

    dd = {"data": data,
          "metadata": metadata}
    return [dd]


def string_to_dtype(astring, dtype):
    astring = astring.strip()
    if dtype == bool:
        return astring.lower() == "true"
    elif dtype in [float, int]:
        return dtype(astring)
    else:
        raise ValueError("No conversion rule for dtype '{}'!".format(dtype))


recipe_tab = {
    "loader": load_tab,
    "suffix": ".tab",
    "mode": "force-distance",
    "maker": "afmformats",
}



"""Test .h5 format writing"""
import pathlib
import tempfile

import h5py

import afmformats


datadir = pathlib.Path(__file__).resolve().parent / "data"


def test_export_hdf5():
    """Test for hotfixes in 0.7.1"""
    jpkfile = datadir / "spot3-0192.jpk-force"
    fdat = afmformats.load_data(jpkfile, mode="force-distance")[0]
    _, path = tempfile.mkstemp(suffix=".h5", prefix="afmformats_test_")

    with h5py.File(path, "w") as h5:
        fdat.export(h5, metadata=True, fmt="hdf5")
        fdat.export(h5, metadata=True, fmt="hdf5")

    with h5py.File(path, "r") as h5:
        assert h5["0"].attrs["enum"] == 0
        assert h5["1"].attrs["enum"] == 1
        assert h5["0"]["force"].attrs["unit"] == "N"
    # cleanup
    try:
        pathlib.Path(path).unlink()
    except OSError:
        pass


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Methods to open JPK data files and to obtain meta data"""
import pathlib
import shutil
import zipfile

import numpy as np

from . import read_jpk_meta as meta

from ..errors import FileFormatNotSupportedError


class ReadJPKError(FileFormatNotSupportedError):
    pass


class ReadJPKColumnError(BaseException):
    pass


extensions = [".jpk-force", ".jpk-force-map"]
VALID_IDS_FORCE = ["vDeflection"]
VALID_IDS_HEIGHT = ["strainGaugeHeight", "capacitiveSensorHeight",
                    "measuredHeight"]
VALID_IDS_PIEZO = ["height", "head-height"]


def find_column(loc_list, id_list):
    """Find a column in a list of strings

    Parameters
    ----------
    loc_list: list of str
        The segment's data location list (within the archive), e.g.

        ['segments/0/channels/height.dat',
         'segments/0/channels/vDeflection.dat',
         'segments/0/channels/strainGaugeHeight.dat']

    id_list: list of str
        Data identifiers to match up, e.g. `VALID_IDS_HEIGHT`

    Returns
    -------
    column: str
        Matched column name, e.g. "strainGaugeHeight"
    loc: str
        Matched segment location,
        e.g. "segments/0/channels/strainGaugeHeight.dat"
    """
    col = None
    for dd in loc_list:
        cn = dd.rsplit("/")[-1].rsplit(".")[0]
        for vc in id_list:
            if vc == cn:
                col = vc
                break
        if col:
            break
    else:
        msg = "No data found for: {}".format(id_list)
        raise ReadJPKError(msg)
    return col, dd


def load_jpk(path, callback=None):
    """Extracts force, measured height, and time from JPK files

    All columns are returned in SI units.

    Parameters
    ----------
    path: str
        Path to a JPK force file
    callback: callable or None
        A method that accepts a float between 0 and 1
        to externally track the process of loading the data.
    """
    path = pathlib.Path(path)
    if callback:
        callback(0)
    # First, only extract the properties
    tdir = meta.extract_jpk(path, props_only=True)
    # Get data file names
    indexlist = meta.get_data_list(path)

    measurements = []
    try:
        with zipfile.ZipFile(str(path)) as arc:
            for ii, item in enumerate(indexlist):
                mm = []
                segroot = tdir / item[0][0].rsplit("segments", 1)[0]
                # go through the segments
                for mi, curve in enumerate(item):
                    if segroot == tdir:
                        enum = 0
                    else:
                        enum = int(segroot.name)
                    # mi == 0: approach
                    # mi == 1: retract
                    # get meta data
                    # (segfolder contains "segment-header.properties")
                    segfolder = tdir / curve[0].rsplit("channels")[0]
                    try:
                        mdi = meta.get_meta_data_seg(segfolder)
                    except meta.ReadJPKMetaKeyError as exc:
                        exc.args = ("{}, File: '{}'".format(exc.args[0], path))
                        raise
                    mdi["enum"] = enum
                    segment = {}
                    # segment time
                    segment["time"] = np.linspace(0,
                                                  mdi["duration"],
                                                  mdi["point count"],
                                                  endpoint=False)
                    # load force data
                    force_col, force_dat = find_column(loc_list=curve,
                                                       id_list=VALID_IDS_FORCE)
                    arc.extract(force_dat, str(tdir))
                    force, unit, _n = load_jpk_single_curve(segroot,
                                                            segment=mi,
                                                            column=force_col,
                                                            slot="force")
                    if unit != "N":
                        msg = "Unknown unit for force: {}".format(unit)
                        raise ReadJPKError(msg)
                    segment["force"] = force

                    # load height (measured) data
                    meas_col, meas_dat = find_column(loc_list=curve,
                                                     id_list=VALID_IDS_HEIGHT)
                    arc.extract(meas_dat, str(tdir))
                    height, unit, _ = load_jpk_single_curve(segroot,
                                                            segment=mi,
                                                            column=meas_col,
                                                            slot="nominal")
                    if unit != "m":
                        msg = "Unknown unit for height: {}".format(unit)
                        raise ReadJPKError(msg)
                    segment["height (measured)"] = height

                    # load height (piezo) data
                    piezo_col, piezo_dat = find_column(loc_list=curve,
                                                       id_list=VALID_IDS_PIEZO)
                    arc.extract(piezo_dat, str(tdir))
                    heightp, unit, _ = load_jpk_single_curve(segroot,
                                                             segment=mi,
                                                             column=piezo_col,
                                                             slot="calibrated")
                    if unit != "m":
                        msg = "Unknown unit for piezo height: {}".format(unit)
                        raise ReadJPKError(msg)
                    segment["height (piezo)"] = heightp

                    mm.append([segment, mdi, path])
                if callback:
                    # Callback with a float between 0 and 1 to update
                    # a progress dialog or somesuch.
                    callback(ii/len(indexlist))

                measurements.append(mm)
                shutil.rmtree(str(segroot))
    except BaseException:
        raise
    finally:
        shutil.rmtree(str(tdir), ignore_errors=True)

    return measurements


def load_jpk_single_curve(path_jpk, segment=0, column="vDeflection",
                          slot="default"):
    """Load a single curve from a jpk-force file

    Parameters
    ----------
    path_jpk : str
        Path to a jpk-force file or to a directory containing "segments"
    segment: int
        Index of the segment to use.
    column: str
        Column name; one of:

            - "height" : piezzo height
            - "vDeflection": measured deflection
            - "straingGaugeHeight": measured height

    slot: str
        The .dat files in the JPK measurement zip files come with different
        calibration slots. Valid values are


            - For the height of the piezo crystal during measurement
              (the piezo height is not as accurate as the measured height
              from the height sensor; the piezo movement is not linear):
              "height.dat": "volts", "nominal", "calibrated"

            - For the measured height of the cantilever:
              "strainGaugeHeight.dat": "volts", "nominal", "absolute"
              "measuredHeight.dat":  "volts", "nominal", "absolute"
              "capacitiveSensorHeight": "volts", "nominal", "absolute"
              (they are all the same)

            - For the recorded cantilever deflection:
              "vDeflection.dat": "volts", "distance", "force"


    Returns
    -------
    data: 1d ndarray
        A numpy array containing the scaled data.
    unit: str
        A string representing the metric unit of the data.
    name: str
        The name of the data column.


    Notes
    -----
    This method does is not designed for directly opening JPK files.
    Please use the `load_jpk` method instead, which wraps around this
    method and handles exceptions better.
    """
    path_jpk = pathlib.Path(path_jpk)
    if path_jpk.is_dir():
        tdir = path_jpk
        cleanup = False
    else:
        tdir = meta.extract_jpk(path_jpk)
        cleanup = True
    segroot = tdir / "segments"
    if not segroot.exists():
        raise OSError("No `segments` subdir found in {}!".format(tdir))
    if not (segroot / str(segment)).exists():
        raise ValueError("Segment {} not found in {}!".format(segment,
                                                              path_jpk))
    chroot = segroot / str(segment) / "channels"
    channels = chroot.glob("*.dat")
    for ch in channels:
        key = ch.stem
        if key == column:
            data = load_dat_unit(ch, slot=slot)
            break
    else:
        msg = "No data for column '{}' and slot '{}'".format(column, slot)
        raise ReadJPKColumnError(msg)
    if cleanup:
        shutil.rmtree(str(tdir))

    return data


def retrieve_segments_data(path_dir):
    """From an extracted jpk file, retrieve the containing segments

    This is a convenience method that returns a list of the measurement
    data with the default slot, including units and column names.

    Parameters
    ----------
    path_dir: str
        Path to a directory containing a "segments" folder.

    Returns
    -------
    segment_list: list
        A list with items: [data, unit, column_name]

    """
    path_dir = pathlib.Path(path_dir)
    segroot = path_dir / "segments"
    segment_data = []
    for se in sorted(segroot.glob("[0-1]")):
        chan_data = []
        chroot = se / "channels"
        for ch in chroot.glob("*.dat"):
            chan_data.append(load_dat_unit(ch))
        segment_data.append(chan_data)
    return segment_data


def load_dat_raw(path_dat):
    """Load data from binary JPK .dat files

    Parameters
    ----------
    path_dat: str
        Path to a .dat file. A `segment-header.properties`
        file must be present in the parent folder.

    Returns
    -------
    data: 1d ndarray
        A numpy array with the raw data.


    Notes
    -----
    This method tries to correctly determine the data type of the
    binary data and scales it with the `data.encoder.scaling`
    values given in the header files.

    See Also
    --------
    load_dat_unit: Includes conversion to useful units

    """
    path_dat = pathlib.Path(path_dat).resolve()
    key = path_dat.stem

    # open header file
    header_file = path_dat.parents[1] / "segment-header.properties"
    prop = meta.get_seg_head_prop(header_file)

    # extract multiplier and offset from header
    # multiplier
    mult_str1 = "channel.{}.data.encoder.scaling.multiplier".format(key)
    mult_str2 = "channel.{}.encoder.scaling.multiplier".format(key)
    try:
        mult = prop[mult_str1]
    except BaseException:
        mult = prop[mult_str2]
    # offset
    off_str1 = "channel.{}.data.encoder.scaling.offset".format(key)
    off_str2 = "channel.{}.encoder.scaling.offset".format(key)
    try:
        off = prop[off_str1]
    except BaseException:
        off = prop[off_str2]
    # get encoder
    enc_str1 = "channel.{}.data.encoder.type".format(key)
    enc_str2 = "channel.{}.encoder.type".format(key)
    try:
        enc = prop[enc_str1]
    except BaseException:
        enc = prop[enc_str2]
    # determine encoder
    if enc == "signedshort":
        mydtype = np.dtype(">i2")
    elif enc == "unsignedshort":
        mydtype = np.dtype(">u2")
    elif enc == "signedinteger":
        mydtype = np.dtype(">i4")
    elif enc == "unsignedinteger":
        mydtype = np.dtype(">u4")
    elif enc == "signedlong":
        mydtype = np.dtype(">i8")
    else:
        raise NotImplementedError("Data file format '{}' not supported".
                                  format(enc))

    data = np.fromfile(str(path_dat), dtype=mydtype) * mult + off
    return data


def load_dat_unit(path_dat, slot="default"):
    """Load data from a JPK .dat file with a specific calibration slot


    Parameters
    ----------
    path_dat : str
        Path to a .dat file
    slot: str
        The .dat files in the JPK measurement zip files come with different
        calibration slots. Valid values are

            - For the height of the piezo crystal during measurement
              (the piezo height is not as accurate as the measured height
              from the height sensor; the piezo movement is not linear):
              "height.dat": "volts", "nominal", "calibrated"

            - For the measured height of the cantilever:
              "strainGaugeHeight.dat": "volts", "nominal", "absolute"
              "measuredHeight.dat":  "volts", "nominal", "absolute"
              "capacitiveSensorHeight": "volts", "nominal", "absolute"
              (they are all the same)

            - For the recorded cantilever deflection:
              "vDeflection.dat": "volts", "distance", "force"


    Returns
    -------
    data: 1d ndarray
        A numpy array containing the scaled data.
    unit: str
        A string representing the metric unit of the data.
    name: str
        The name of the data column.


    Notes
    -----
    The raw data (see `load_dat_raw`) is usually stored in "volts" and
    needs to be converted to e.g. "force" for "vDeflection" or "nominal"
    for "strainGaugeHeight". The conversion parameters (offset, multiplier)
    are stored in the header files and they are not stored separately for
    each slot, but the conversion parameters are stored relative to the
    slots. For instance, to compute the "force" slot from the raw "volts"
    data, one first needs to compute the "distance" slot. This conversion
    is taken care of by this method.

    This is an example header:

        channel.vDeflection.data.file.name=channels/vDeflection.dat
        channel.vDeflection.data.file.format=raw
        channel.vDeflection.data.type=short
        channel.vDeflection.data.encoder.type=signedshort
        channel.vDeflection.data.encoder.scaling.type=linear
        channel.vDeflection.data.encoder.scaling.style=offsetmultiplier
        channel.vDeflection.data.encoder.scaling.offset=-0.00728873489143207
        channel.vDeflection.data.encoder.scaling.multiplier=3.0921021713588157E-4
        channel.vDeflection.data.encoder.scaling.unit.type=metric-unit
        channel.vDeflection.data.encoder.scaling.unit.unit=V
        channel.vDeflection.channel.name=vDeflection
        channel.vDeflection.conversion-set.conversions.list=distance force
        channel.vDeflection.conversion-set.conversions.default=force
        channel.vDeflection.conversion-set.conversions.base=volts
        channel.vDeflection.conversion-set.conversion.volts.name=Volts
        channel.vDeflection.conversion-set.conversion.volts.defined=false
        channel.vDeflection.conversion-set.conversion.distance.name=Distance
        channel.vDeflection.conversion-set.conversion.distance.defined=true
        channel.vDeflection.conversion-set.conversion.distance.type=simple
        channel.vDeflection.conversion-set.conversion.distance.comment=Distance
        channel.vDeflection.conversion-set.conversion.distance.base-calibration-slot=volts
        channel.vDeflection.conversion-set.conversion.distance.calibration-slot=distance
        channel.vDeflection.conversion-set.conversion.distance.scaling.type=linear
        channel.vDeflection.conversion-set.conversion.distance.scaling.style=offsetmultiplier
        channel.vDeflection.conversion-set.conversion.distance.scaling.offset=0.0
        channel.vDeflection.conversion-set.conversion.distance.scaling.multiplier=7.000143623002982E-8
        channel.vDeflection.conversion-set.conversion.distance.scaling.unit.type=metric-unit
        channel.vDeflection.conversion-set.conversion.distance.scaling.unit.unit=m
        channel.vDeflection.conversion-set.conversion.force.name=Force
        channel.vDeflection.conversion-set.conversion.force.defined=true
        channel.vDeflection.conversion-set.conversion.force.type=simple
        channel.vDeflection.conversion-set.conversion.force.comment=Force
        channel.vDeflection.conversion-set.conversion.force.base-calibration-slot=distance
        channel.vDeflection.conversion-set.conversion.force.calibration-slot=force
        channel.vDeflection.conversion-set.conversion.force.scaling.type=linear
        channel.vDeflection.conversion-set.conversion.force.scaling.style=offsetmultiplier
        channel.vDeflection.conversion-set.conversion.force.scaling.offset=0.0
        channel.vDeflection.conversion-set.conversion.force.scaling.multiplier=0.043493666407368466
        channel.vDeflection.conversion-set.conversion.force.scaling.unit.type=metric-unit
        channel.vDeflection.conversion-set.conversion.force.scaling.unit.unit=N

    To convert from the raw "volts" data to force data, these steps are
    performed:

    - Convert from "volts" to "distance" first, because the
      "base-calibration-slot" for force is "distance".

      distance = volts*7.000143623002982E-8 + 0.0

    - Convert from "distance" to "force":

      force = distance*0.043493666407368466 + 0.0

    The multipliers shown above are the values for sensitivity and spring
    constant:
    sensitivity = 7.000143623002982E-8 m/V
    spring_constant = 0.043493666407368466 N/m
    """
    path_dat = pathlib.Path(path_dat).resolve()
    key = path_dat.stem

    # open header file
    header_file = path_dat.parents[1] / "segment-header.properties"
    prop = meta.get_seg_head_prop(header_file)

    conv = "channel.{}.conversion-set".format(key)

    if slot == "default":
        slot = prop[conv+".conversions.default"]

    # get base unit
    base = prop[conv+".conversions.base"]

    # Now iterate through the conversion sets until we have the base converter.
    # A list of multipliers and offsets
    converters = []
    curslot = slot

    while curslot != base:
        # Get current slot multipliers and offsets
        off_str = conv+".conversion.{}.scaling.offset".format(curslot)
        off = prop[off_str]
        mult_str = conv+".conversion.{}.scaling.multiplier".format(curslot)
        mult = prop[mult_str]
        converters.append([mult, off])
        sl_str = conv+".conversion.{}.base-calibration-slot".format(curslot)
        curslot = prop[sl_str]

    # Get raw data
    data = load_dat_raw(path_dat)
    for c in converters[::-1]:
        data[:] = c[0] * data[:] + c[1]

    if base == slot:
        unit_str = "channel.{}.data.encoder.scaling.unit.unit".format(key)
        unit = prop[unit_str]
    else:
        try:
            unit_str = conv+".conversion.{}.scaling.unit".format(slot)
            unit = prop[unit_str]
        except KeyError:
            unit_str = conv+".conversion.{}.scaling.unit.unit".format(slot)
            unit = prop[unit_str]

    name_str = conv+".conversion.{}.name".format(slot)
    name = prop[name_str]
    return data, unit, "{} ({})".format(key, name)



from . import meta  # noqa: 401
from .formats import load_data  # noqa: 401
from .formats import supported_extensions  # noqa: 401
from .afm_fdist import AFMForceDistance  # noqa: 401

from ._version import version as __version__  # noqa: 401



"""Test of basic opening functionalities"""
import pathlib
import shutil

from afmformats.fmt_jpk import read_jpk, read_jpk_meta


datadir = pathlib.Path(__file__).resolve().parent / "data"


def test_open_jpk_simple():
    jpkfile = datadir / "spot3-0192.jpk-force"
    tdir = read_jpk_meta.extract_jpk(jpkfile)
    segroot = tdir / "segments"
    seg = sorted(segroot.glob("[0-1]"))
    chan_data = {}
    chroot = seg[0] / "channels"
    for ch in chroot.glob("*.dat"):
        key = ch.stem
        chan_data[key] = read_jpk.load_dat_raw(ch)
    assert chan_data["height"][0] == 50.425720226584403
    shutil.rmtree(tdir, ignore_errors=True)


def test_open_jpk_calibration():
    cf = datadir / "calibration_force-save-2015.02.04-11.25.21.294.jpk-force"
    try:
        read_jpk.load_jpk(cf)
    except read_jpk_meta.ReadJPKMetaKeyError:
        pass
    else:
        assert False, "no spring constant should raise error"


def test_open_jpk_conversion():
    jpkfile = datadir / "spot3-0192.jpk-force"
    tdir = read_jpk_meta.extract_jpk(jpkfile)
    segroot = tdir / "segments"
    seg = sorted(segroot.glob("[0-1]"))
    chan_data = {}
    chroot = segroot / seg[0] / "channels"
    for ch in chroot.glob("*.dat"):
        key = ch.stem
        chan_data[key] = read_jpk.load_dat_unit(ch)

    assert chan_data["vDeflection"][2] == "vDeflection (Force)"
    assert chan_data["vDeflection"][1] == "N"
    assert chan_data["vDeflection"][0][0] == -5.145579192349918e-10
    assert chan_data["height"][0][0] == 2.8783223430683289e-05
    assert chan_data["strainGaugeHeight"][0][0] == 2.2815672438768612e-05
    shutil.rmtree(tdir, ignore_errors=True)


def test_get_single_curves():
    jpkfile = datadir / "spot3-0192.jpk-force"
    read_jpk.load_jpk_single_curve(jpkfile, column="vDeflection", slot="force")
    read_jpk.load_jpk_single_curve(jpkfile, column="height")
    read_jpk.load_jpk_single_curve(
        jpkfile, column="strainGaugeHeight", slot="nominal")
    read_jpk.load_jpk_single_curve(jpkfile, column="height", slot="volts")
    read_jpk.load_jpk_single_curve(
        jpkfile, segment=1, column="height", slot="volts")
    # This is height in the txt files
    a = read_jpk.load_jpk_single_curve(
        jpkfile, column="height", slot="nominal")
    assert a[0][0] == 4.9574279773415606e-05


def test_meta():
    jpkfile = datadir / "spot3-0192.jpk-force"
    tdir = read_jpk_meta.extract_jpk(jpkfile)
    sdir = tdir / "segments" / "0"
    md = read_jpk_meta.get_meta_data_seg(sdir)
    assert md["spring constant"] == 0.043493666407368466


def test_load_jpk():
    jpkfile = datadir / "spot3-0192.jpk-force"
    segs = read_jpk.load_jpk(jpkfile)
    tdir = read_jpk_meta.extract_jpk(jpkfile)
    sdir = tdir / "segments" / "0"
    md = read_jpk_meta.get_meta_data_seg(sdir)
    assert md["imaging mode"] == "force-distance"
    assert len(segs) == 1, "Only one measurement"
    assert len(segs[0]) == 2, "approach and retract curves"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test of basic opening functionalities"""
import pathlib

import numpy as np

import afmformats


datadir = pathlib.Path(__file__).resolve().parent / "data"


def test_load_jpk_map():
    jpkfile = datadir / "map2x2_extracted.jpk-force-map"
    afmlist = afmformats.load_data(path=jpkfile)

    assert len(afmlist) == 4
    ds = afmlist[2]
    ds2 = afmlist[3]
    assert ds.metadata["enum"] == 2
    assert np.allclose(ds.metadata["duration"], 1.6089775561097257)
    # Verified with visual inspection of force curve in JPK software
    assert np.allclose(ds["force"][0], -5.8540192943834714e-10)
    assert np.allclose(ds["height (measured)"][0], 0.0001001727719556085)
    # make sure curve id is different
    assert ds.metadata["curve id"] != ds2.metadata["curve id"]
    assert ds.metadata["session id"] == ds2.metadata["session id"]


def test_load_jpk_simple():
    jpkfile = datadir / "spot3-0192.jpk-force"
    afmlist = afmformats.load_data(path=jpkfile)
    ds = afmlist[0]
    assert ds.metadata["enum"] == 0
    assert np.allclose(ds["height (measured)"][0], 2.2815672438768612e-05)


def test_load_jpk_piezo():
    jpkfile = datadir / "spot3-0192.jpk-force"
    afmlist = afmformats.load_data(path=jpkfile)
    ds = afmlist[0]
    assert np.allclose(ds["height (piezo)"][0], 2.878322343068329e-05)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib

import numpy as np

from .parse_funcs import fint, vd_str_in


__all__ = ["META_FIELDS", "DEF_ALL", "KEYS_VALID", "MetaDataMissingError",
           "MetaData"]


#: Compedium of all allowed meta data keys, sorted by topic, and
#: including units and validation methods
META_FIELDS = {
    # AFM acquisition settings
    "acquisition": {
        "feedback mode": ["Feedback mode", "", vd_str_in([
            # JPK contact mode
            "contact",
            # From the NanoWizard User Manual (v. 4.2) sec. 5.7:
            # Force modulation mode is a mixture between contact mode and
            # intermittend mode and "can be thought of as a kind of contact
            # mode with an added vibration on the cantilever".
            "force-modulation",
        ])],
        "imaging mode": ["Imaging modality", "", vd_str_in([
            # force-distance measurements
            "force-distance",
        ])],
        "rate approach": ["Sampling rate (approach)", "Hz", float],
        "rate retract": ["Sampling rate (retract)", "Hz", float],
        "sensitivity": ["Sensitivity", "m/V", float],
        "setpoint": ["Active feedback loop setpoint", "N", float],
        "spring constant": ["Cantilever spring constant", "N/m", float],
        "z range": ["Axial piezo range covered", "m", float],
    },
    # dataset parameters
    "dataset": {
        "duration": ["Duration", "s", float],
        "enum": ["Dataset index within the experiment", "", fint],
        "point count": ["Size of the dataset in points", "", fint],
        "speed approach": ["Piezo speed (approach)", "m/s", float],
        "speed retract": ["Piezo speed (retract)", "m/s", float],
    },
    # QMap related dataset metadata
    "qmap": {
        "grid center x": ["Horizontal center of grid", "m", float],
        "grid center y": ["Vertical center of grid", "m", float],
        "grid index x": ["Horizontal grid position index", "", fint],
        "grid index y": ["Vertical grid position index", "", fint],
        "grid shape x": ["Horizontal grid shape", "px", fint],
        "grid shape y": ["Vertical grid shape", "px", fint],
        "grid size x": ["Horizontal grid size", "m", float],
        "grid size y": ["Vertical grid size", "m", float],
        "position x": ["Horizontal position", "m", float],
        "position y": ["Vertical position", "m", float],
    },
    # AFM setup
    "setup": {
        "instrument": ["Instrument", "", str],
        "software": ["Acquisition software", "", str],
        "software version": ["Acquisition software version", "", str],
    },
    # storage data
    "storage": {
        "curve id": ["Curve identifier", "", str],
        "date": ["Recording date", "", str],  # YYYY-MM-DD
        "path": ["Path", "", pathlib.Path],
        "session id": ["Dataset identifier", "", str],
        "time": ["Recording time", "", str],  # HH:MM:SS.S
    },
}


#: A dictionary for all metadata definitions
DEF_ALL = {}
for _sec in META_FIELDS:
    for _key in META_FIELDS[_sec]:
        DEF_ALL[_key] = META_FIELDS[_sec][_key]

#: List of all valid meta data keys
KEYS_VALID = sorted(DEF_ALL.keys())


class MetaDataMissingError(BaseException):
    """Raised when meta data is missing"""
    pass


class MetaData(dict):
    """Management of meta data variables

    Valid key names are defined in :const:`afmformats.meta.KEYS_VALID`.
    """
    valid_keys = KEYS_VALID

    def __init__(self, *args, **kwargs):
        super(MetaData, self).__init__(*args, **kwargs)
        # check for invalid keys
        for key in self:
            if key not in self.valid_keys:
                raise KeyError("Unknown metadata key: '{}'".format(key))

    def __setitem__(self, key, value):
        """Set a metadata key

        The key must be a valid key defined in `self.valid_keys`
        (defaults to :const:`afmformats.meta.KEYS_VALID`).
        """
        if key not in self.valid_keys:
            raise KeyError("Unknown metadata key: '{}'".format(key))
        super(MetaData, self).__setitem__(key, value)

    def __getitem__(self, *args, **kwargs):
        if args[0] not in self and args[0] in self.valid_keys:
            msg = "No meta data was defined for '{}'!".format(args[0]) \
                  + "Please make sure you passed the dictionary `metadata` " \
                  + "when you loaded your data."
            raise MetaDataMissingError(msg)
        elif args[0] not in self:
            msg = "Unknown meta key: '{}'!".format(args[0])
            raise KeyError(msg)
        return super(MetaData, self).__getitem__(*args, **kwargs)

    def get_summary(self):
        """Convenience function returning the meta data summary

        Returns a dict of dicts with keys matching the DEF_* dicts.
        Unset values are returned as `np.nan`.
        """
        summary = {}
        for sec in META_FIELDS:
            summary[sec] = {}
            for key in META_FIELDS[sec]:
                summary[sec][key] = self.get(key, np.nan)
        return summary



from .afm_data import AFMData


__all__ = ["AFMForceDistance", "Segment"]


class AFMForceDistance(AFMData):
    """Base class for AFM force-distance data

    A force-distance dataset consists of an approach and
    a retract curve.
    """

    def __repr__(self):
        repre = "<Force-Distance Data '{}'[{}] at {}>".format(
            self.path, self.enum, hex(id(self)))
        return repre

    def __setitem__(self, key, values):
        if len(values) != len(self):
            raise ValueError(
                "Cannot set data of length '{}' ".format(len(values))
                + "for AFMForceDistance of length '{}'!".format(len(self)))
        # do not touch raw data
        self._data[key] = values

    @property
    def appr(self):
        """Dictionary-like interface to the approach segment"""
        return Segment(self._raw_data, self._data, which="approach")

    @property
    def mode(self):
        """Imaging modality"""
        return "force-distance"

    @property
    def retr(self):
        """Dictionary-like interface to the retract segment"""
        return Segment(self._raw_data, self._data, which="retract")


class Segment(object):
    """Simple wrapper around dict-like `data` to expose a single segment"""

    def __init__(self, raw_data, data, which="approach"):
        if which not in ["approach", "retract"]:
            raise ValueError("`which` must be 'approach' or 'retract', "
                             + "got '{}'!".format(which))
        #: The segment type (approach or retract)
        self.which = which
        if which == "approach":
            self.idx = False
        else:
            self.idx = True
        self.raw_data = raw_data
        self.data = data

    def __getitem__(self, key):
        used = self.data["segment"] == self.idx
        if key in self.data:
            return self.data[key][used]
        elif key in self.raw_data:
            return self.raw_data[key][used].copy()
        else:
            raise KeyError("Undefined feature '{}'!".format(key))



"""Test of basic opening functionalities"""
import pathlib

import numpy as np

import afmformats
from afmformats.fmt_jpk import read_jpk


datadir = pathlib.Path(__file__).resolve().parent / "data"


def test_open_jpk_map():
    jpkfile = datadir / "map0d_extracted.jpk-force-map"
    data = read_jpk.load_jpk(jpkfile)
    force = data[0][0][0]["force"]
    height = data[0][0][0]["height (measured)"]
    # Verified with visual inspection of force curve in JPK software
    assert np.allclose(force[0], -4.7426862623854873e-10)
    assert np.allclose(height[0], 7.0554296897149161e-05)


def test_open_jpk_map_enum():
    jpkfile = datadir / "map2x2_extracted.jpk-force-map"
    data = afmformats.load_data(jpkfile)
    # Correct enumeration
    assert data[0].metadata["enum"] == 0
    assert data[1].metadata["enum"] == 1
    assert data[2].metadata["enum"] == 2
    assert data[3].metadata["enum"] == 3


def test_open_jpk_map2():
    jpkfile = datadir / "map2x2_extracted.jpk-force-map"
    data = read_jpk.load_jpk(jpkfile)
    force = data[2][0][0]["force"]
    height = data[2][0][0]["height (measured)"]
    assert len(data) == 4
    assert np.allclose(force[0], -5.8540192943834714e-10)
    assert np.allclose(height[0], 0.0001001727719556085)
    assert data[0][0][1]["grid index x"] == 0
    assert data[1][0][1]["grid index x"] == 9
    assert data[2][0][1]["grid index x"] == 9
    assert data[3][0][1]["grid index x"] == 0
    assert data[3][0][1]["grid index x"] == data[3][1][1]["grid index x"]


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Methods to open JPK data files and to obtain meta data"""
import functools
import pathlib
import tempfile
import zipfile

import jprops
import numpy as np

from ..errors import FileFormatMetaDataError
from .. import meta


class ReadJPKMetaKeyError(FileFormatMetaDataError):
    pass


def extract_jpk(path_jpk, props_only=False):
    """Extract the JPK data files and return the extracted path"""
    tdir = tempfile.mkdtemp(prefix="afmformats_jpk_")
    with zipfile.ZipFile(str(path_jpk)) as fd:
        if props_only:
            for name in fd.namelist():
                if name.endswith(".properties"):
                    fd.extract(name, tdir)
        else:
            fd.extractall(tdir)
    return pathlib.Path(tdir)


def isnan(obj):
    if isinstance(obj, (str)):
        if obj.lower() == "nan":
            return True
    elif isinstance(obj, float):
        return np.isnan(obj)
    else:
        return False


def get_data_list(path_jpk):
    path_jpk = pathlib.Path(path_jpk)
    with zipfile.ZipFile(str(path_jpk)) as fd:
        files = fd.namelist()
    dfiles = [f for f in files if f.endswith(".dat")]
    # sort files correctly
    segdict = {}
    if "segments/" in files:
        app = [aa for aa in dfiles if aa.startswith("segments/0")]
        ret = [aa for aa in dfiles if aa.startswith("segments/1")]
        segdict["0"] = (app, ret)
    elif "index/" in files:
        segids = [dd.split("/")[1] for dd in dfiles]
        segids = list(set(segids))
        for sid in segids:
            sidstrapp = "index/{}/segments/0".format(sid)
            app = [dd for dd in dfiles if dd.startswith(sidstrapp)]
            sidstrret = "index/{}/segments/1".format(sid)
            ret = [dd for dd in dfiles if dd.startswith(sidstrret)]
            segdict[sid] = (app, ret)
    else:
        msg = "Unknown JPK file format: {}".format(path_jpk)
        raise NotImplementedError(msg)

    segkeys = list(segdict.keys())

    segkeys.sort(key=lambda x: int(x))

    seglist = []
    for key in segkeys:
        seglist.append(segdict[key])
    return seglist


def get_meta_data_seg(path_segment):
    """Obtain most important meta-data for a segment folder
    """
    segment = pathlib.Path(path_segment)
    conv = "channel.vDeflection.conversion-set.conversion"
    sens_str = conv + ".distance.scaling.multiplier"
    sprc_str = conv + ".force.scaling.multiplier"
    pos_str = "force-segment-header.environment.xy-scanner-position-map"\
              + ".xy-scanner.motorstage.position."

    # These are properties that will be returned, if they exist in the header
    # files.
    dvars = [
        # acquisition
        ["sensitivity", sens_str],
        ["spring constant", sprc_str],
        ["feedback mode", "force-segment-header.settings.feedback-mode.name"],
        # dataset
        ["duration", "force-segment-header.duration"],
        ["point count", "force-segment-header.num-points"],
        # qmap
        ["grid shape x", "force-scan-map.position-pattern.grid.ilength"],
        ["grid shape y", "force-scan-map.position-pattern.grid.jlength"],
        # storage
        ["session id", "force-segment-header.approach-id"],
        # setup
        ["instrument", "force-scan-series.description.instrument"],
        ["software version", "force-scan-series.description.source-software"],
    ]

    # These are properties that will not be returned, but are used for the
    # computation of properties.
    dvars_im = [
        ["z start", "force-segment-header.settings.segment-settings.z-start"],
        ["z end", "force-segment-header.settings.segment-settings.z-end"],
        # Setpoint is stored in Volts and needs to be converted still
        ["setpoint [V]",
         "force-segment-header.settings.segment-settings.setpoint"],
        ["position x [m]", pos_str + "x"],
        ["position y [m]", pos_str + "y"],
        ["grid size x [m]", "force-scan-map.position-pattern.grid.ulength"],
        ["grid size y [m]", "force-scan-map.position-pattern.grid.vlength"],
        ["grid center x [m]", "force-scan-map.position-pattern.grid.xcenter"],
        ["grid center y [m]", "force-scan-map.position-pattern.grid.ycenter"],
        ["curve type", "force-segment-header.settings.style"],
        ["position index", "force-scan-series.header.position-index"]
    ]

    md = meta.MetaData()
    # Currently, only force-distance mode is supported!
    md["imaging mode"] = "force-distance"
    if segment.name == "0":
        curseg = "approach"
    else:
        curseg = "retract"
    md["software"] = "JPK"
    md_im = {}
    header_file = segment / "segment-header.properties"
    prop = get_seg_head_prop(header_file)

    # date and time
    md["date"], md["time"], _ = prop["force-segment-header.time-stamp"].split()

    for mdi, dvarsi in [[md, dvars], [md_im, dvars_im]]:
        for name, var in dvarsi:
            if var in prop:
                mdi[name] = prop[var]

    for mkey in ["spring constant",
                 "sensitivity",
                 ]:
        if mkey not in md:
            msg = "Missing meta data: '{}'".format(mkey)
            raise ReadJPKMetaKeyError(msg)

    md["curve id"] = "{}:{:g}".format(md["session id"], mdi["position index"])

    md["setpoint"] = md_im["setpoint [V]"] * \
        md["spring constant"]*md["sensitivity"]

    md["rate " + curseg] = md["point count"]/md["duration"]
    md["z range"] = abs(md_im["z start"] - md_im["z end"])
    md["speed " + curseg] = md["z range"]/md["duration"]
    if "position x [m]" in md_im:
        md["position x"] = md_im["position x [m]"]
    if "position y [m]" in md_im:
        md["position y"] = md_im["position y [m]"]
    if "grid size x [m]" in md_im:
        md["grid size x"] = md_im["grid size x [m]"]
    if "grid size y [m]" in md_im:
        md["grid size y"] = md_im["grid size y [m]"]
    if "grid center x [m]" in md_im:
        md["grid center x"] = md_im["grid center x [m]"]
    if "grid center y [m]" in md_im:
        md["grid center y"] = md_im["grid center y [m]"]
    if ("position x [m]" in md_im and
        "position y [m]" in md_im and
        "grid size x [m]" in md_im and
        "grid size y [m]" in md_im and
        "grid center x [m]" in md_im and
        "grid center y [m]" in md_im and
        "grid shape x" in md and
            "grid shape y" in md):
        pxpx = position_m2px(pos_m=md_im["position x [m]"],
                             size_m=md_im["grid size x [m]"],
                             center_m=md_im["grid center x [m]"],
                             size_px=md["grid shape x"])
        pypx = position_m2px(pos_m=md_im["position y [m]"],
                             size_m=md_im["grid size y [m]"],
                             center_m=md_im["grid center y [m]"],
                             size_px=md["grid shape y"])
        md["grid index x"] = pxpx
        md["grid index y"] = pypx

    integer_keys = ["grid shape x",
                    "grid shape y",
                    "grid index x",
                    "grid index y",
                    "point count",
                    ]
    for ik in integer_keys:
        if ik in md:
            md[ik] = int(round(md[ik]))
    return md


@functools.lru_cache(maxsize=10)
def get_seg_head_prop(path_seg_head_prop):
    """ Obtain the properies of a "segment-header.properties" file

    Parameters
    ----------
    path_seg_head_prop : str
        Full path to a "segment-header.properties" file

    Notes
    -----
    This method also parses these files if present:
    - "shared-data/header.properties"
    - "header.properties"
    """
    path = pathlib.Path(path_seg_head_prop).resolve()
    with path.open(mode="rb") as fd:
        prop = jprops.load_properties(fd)

    # Determine if we have a shared-data file
    # These are positions in the file system that could contain the
    # shared properties files:
    shared_locs = [path.parents[2] / "shared-data",
                   path.parents[4] / "shared-data",
                   ]
    general_locs = [path.parents[2],
                    path.parents[4]
                    ]

    # Loop through the candidates
    for cc in shared_locs:
        shared = cc / "header.properties"
        if shared.exists():
            # A candidate exists. Load its properties.
            psprop = load_prop_file(shared)
            # Generate lists of keys and sort them for easier debugging.
            proplist = list(prop.keys())
            proplist.sort()
            pslist = list(psprop.keys())
            pslist.sort()
            # Loop through the segment data and search for lcd-info tags
            for key in proplist:
                # Get line channel data
                if key.count(".*"):
                    # Replace the lcd-info tag by the values in the shared
                    # properties file:
                    # 0, 1, 2, 3, etc.
                    index = prop[key]
                    # lcd-info, force-segment-header-info
                    mediator = ".".join(key.split(".")[-2:-1])
                    # channel.vDeflection, force-segment-header
                    headkey = key.rsplit(".", 2)[0]
                    # append a "." here to make sure
                    # not to confuse "1" with "10".
                    startid = "{}.{}.".format(mediator, index)

                    for k2 in pslist:
                        if k2.startswith(startid):
                            var = ".".join(k2.split(".")[2:])
                            prop[".".join([headkey, var])] = psprop[k2]

    for gg in general_locs:
        gen = gg / "header.properties"
        if gen.exists():
            gsprop = load_prop_file(gen)
            # Add all other keys
            for pk in gsprop:
                prop[pk] = gsprop[pk]

    for p in prop:
        try:
            prop[p] = float(prop[p])
        except BaseException:
            pass

    return prop


@functools.lru_cache(maxsize=10)
def load_prop_file(path):
    path = pathlib.Path(path)
    with path.open(mode="rb") as fd:
        props = jprops.load_properties(fd)
    return props


def position_m2px(pos_m, size_m, center_m, size_px):
    """Convert qmap positions from [m] to array coordinates in [px]

    Parameters
    ----------
    pos_m: float
        positions [m]
    size_m: float
        grid size [m]
    center_m: float
        grid center position [m]
    size_px: int
        grid size [px]

    Returns
    -------
    pos_px: int
        index position of `pos_m`
    """
    if size_px != int(size_px):
        raise ValueError("`size_px` must be integer, got {}!".format(size_px))
    size_px = int(size_px)
    s1 = center_m - size_m / 2
    s2 = center_m + size_m / 2

    x, dx = np.linspace(s1, s2, size_px, endpoint=False, retstep=True)
    x += dx/2

    xpx = np.nanargmin(np.abs(x-pos_m))
    return xpx



import pathlib

import numpy as np
from . import read_jpk


__all__ = ["load_jpk"]


def load_jpk(path, callback=None):
    """Loads JPK Instruments data files

    These files are zip files containing java property files and
    integer-encoded binary data. The property files include recipes
    on how to convert the raw integer data to SI units.
    """
    # load the data
    measurements = read_jpk.load_jpk(path=path, callback=callback)
    # convert join the segment data
    dataset = []
    for mm in measurements:
        (app, metadata, _), (ret, metadata_ret, _) = mm
        metadata["path"] = pathlib.Path(path)
        metadata["duration"] += metadata_ret["duration"]
        metadata["rate retract"] = metadata_ret["rate retract"]
        metadata["point count"] += metadata_ret["point count"]
        metadata["speed retract"] = metadata_ret["speed retract"]
        # join segments
        lenapp = len(app[list(app.keys())[0]])
        lenret = len(ret[list(ret.keys())[0]])
        ret["time"] += metadata["duration"]
        data = {}
        for key in app.keys():
            data[key] = np.concatenate((app[key], ret[key]))
        data["segment"] = np.concatenate((np.zeros(lenapp, dtype=bool),
                                          np.ones(lenret, dtype=bool)))
        dataset.append({"data": data,
                        "metadata": metadata,
                        })
    return dataset


recipe_jpk_force = {
    "loader": load_jpk,
    "suffix": ".jpk-force",
    "mode": "force-distance",
    "maker": "JPK Instruments",
}

recipe_jpk_force_map = {
    "loader": load_jpk,
    "suffix": ".jpk-force-map",
    "mode": "force-distance",
    "maker": "JPK Instruments",
}



class AFMFileFormatError(BaseException):
    pass


class FileFormatNotSupportedError(AFMFileFormatError):
    pass


class FileFormatMetaDataError(AFMFileFormatError):
    pass


class DataFileBrokenError(AFMFileFormatError):
    pass


__all__ = [e for e in dir() if not e.startswith("__")]



"""Test of basic opening functionalities"""
import pathlib
import shutil

import numpy as np

import afmformats
from afmformats.fmt_jpk import read_jpk_meta


datadir = pathlib.Path(__file__).resolve().parent / "data"


def test_meta_extracted_map():
    jpkfile = datadir / "map2x2_extracted.jpk-force-map"
    md = afmformats.load_data(jpkfile)[0].metadata
    summary = md.get_summary()
    for sec in ["acquisition", "dataset", "setup", "storage", "qmap"]:
        for kk in summary[sec]:
            v = summary[sec][kk]
            if not isinstance(v, (str, pathlib.Path)):
                assert not np.isnan(v), "should not be nan: '{}'".format(kk)


def test_meta_extracted_map2():
    jpkfile = datadir / "map0d_extracted.jpk-force-map"
    md = afmformats.load_data(jpkfile)[0].metadata
    summary = md.get_summary()
    for sec in ["acquisition", "dataset", "setup", "storage", "qmap"]:
        for kk in summary[sec]:
            v = summary[sec][kk]
            if not isinstance(v, (str, pathlib.Path)):
                assert not np.isnan(v), "should not be nan: '{}'".format(kk)


def test_meta_extracted_single():
    jpkfile = datadir / "spot3-0192.jpk-force"
    md = afmformats.load_data(jpkfile)[0].metadata
    summary = md.get_summary()
    for sec in ["acquisition", "dataset", "setup", "storage"]:
        for kk in summary[sec]:
            v = summary[sec][kk]
            if not isinstance(v, (str, pathlib.Path)):
                assert not np.isnan(v), "should not be nan: '{}'".format(kk)


def test_meta_simple():
    jpkfilemap = datadir / "map0d_extracted.jpk-force-map"
    tdir = read_jpk_meta.extract_jpk(jpkfilemap)
    hpfile = tdir / "index/0/segments/0/segment-header.properties"
    read_jpk_meta.get_seg_head_prop(hpfile)
    shutil.rmtree(tdir, ignore_errors=True)


def test_meta_single():
    jpkfile = datadir / "spot3-0192.jpk-force"
    tdir = read_jpk_meta.extract_jpk(jpkfile)
    sdir = tdir / "segments" / "0"
    md = read_jpk_meta.get_meta_data_seg(sdir)
    assert md["imaging mode"] == "force-distance"
    assert md["duration"] == 0.9999999999999998
    shutil.rmtree(tdir, ignore_errors=True)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
